from dataclasses import dataclass

from koi_net.protocol.knowledge_object import KnowledgeObject
from koi_net.protocol.event import EventType
from ..interfaces import KnowledgeHandler, STOP_CHAIN, HandlerType
from ..cache import Cache


@dataclass
class BasicManifestHandler(KnowledgeHandler):
    cache: Cache
    
    handler_type = HandlerType.Manifest
    
    def handle(self, kobj: KnowledgeObject):
        """Normalized event decider based on manifest and cache state.
    
        Stops processing for manifests which have the same hash, or aren't 
        newer than the cached version. Sets the normalized event type to 
        `NEW` or `UPDATE` depending on whether the RID was previously known.
        """
        
        prev_bundle = self.cache.read(kobj.rid)

        if prev_bundle:
            if kobj.manifest.sha256_hash == prev_bundle.manifest.sha256_hash:
                self.log.debug("Hash of incoming manifest is same as existing knowledge, ignoring")
                return STOP_CHAIN
            if kobj.manifest.timestamp <= prev_bundle.manifest.timestamp:
                self.log.debug("Timestamp of incoming manifest is the same or older than existing knowledge, ignoring")
                return STOP_CHAIN
            
            self.log.debug("RID previously known to me, labeling as 'UPDATE'")
            kobj.normalized_event_type = EventType.UPDATE

        else:
            self.log.debug("RID previously unknown to me, labeling as 'NEW'")
            kobj.normalized_event_type = EventType.NEW
            
        return kobj


from dataclasses import dataclass

from rid_lib.ext import Bundle
from rid_lib.types import KoiNetNode

from koi_net.exceptions import RequestError
from koi_net.config.base import BaseNodeConfig
from koi_net.protocol.node import NodeProfile, NodeType
from koi_net.protocol.edge import EdgeProfile, EdgeStatus, EdgeType, generate_edge_bundle
from koi_net.protocol.knowledge_object import KnowledgeObject
from ..interfaces import KnowledgeHandler, HandlerType
from ..identity import NodeIdentity
from ..kobj_queue import KobjQueue
from ..cache import Cache
from ..request_handler import RequestHandler
from ..graph import NetworkGraph


@dataclass
class NodeContactHandler(KnowledgeHandler):
    identity: NodeIdentity
    cache: Cache
    config: BaseNodeConfig
    kobj_queue: KobjQueue
    graph: NetworkGraph
    request_handler: RequestHandler
    
    handler_type = HandlerType.Network
    rid_types = (KoiNetNode,)
    
    def handle(self, kobj: KnowledgeObject):
        """Makes contact with providers of RID types of interest.
        
        When an incoming node knowledge object is identified as a provider
        of an RID type of interest, this handler will propose a new edge 
        subscribing to future node events, and fetch existing nodes to catch 
        up to the current state.
        """
        # prevents nodes from attempting to form a self loop
        if kobj.rid == self.identity.rid:
            return
        
        node_profile = kobj.bundle.validate_contents(NodeProfile)
        
        available_rid_types = list(
            set(self.config.koi_net.rid_types_of_interest) & 
            set(node_profile.provides.event)
        )
        
        if not available_rid_types:
            return
        
        edge_rid = self.graph.get_edge(
            source=kobj.rid,
            target=self.identity.rid,
        )
        
        # already have an edge established
        if edge_rid:
            prev_edge_bundle = self.cache.read(edge_rid)
            edge_profile = prev_edge_bundle.validate_contents(EdgeProfile)
            
            if set(edge_profile.rid_types) == set(available_rid_types):
                # no change in rid types
                return
            
            self.log.info(f"Proposing updated edge with node provider {available_rid_types}")
            
            edge_profile.rid_types = available_rid_types
            edge_profile.status = EdgeStatus.PROPOSED
            edge_bundle = Bundle.generate(edge_rid, edge_profile.model_dump())
        
        # no existing edge
        else:
            self.log.info(f"Proposing new edge with node provider {available_rid_types}")
            edge_bundle = generate_edge_bundle(
                source=kobj.rid,
                target=self.identity.rid,
                rid_types=available_rid_types,
                edge_type=(
                    EdgeType.WEBHOOK
                    if self.identity.profile.node_type == NodeType.FULL
                    else EdgeType.POLL
                )
            )
        
        # queued for processing
        self.kobj_queue.push(bundle=edge_bundle)
        
        self.log.info("Catching up on network state")
        try:
            payload = self.request_handler.fetch_rids(
                node=kobj.rid, 
                rid_types=available_rid_types
            )
        except RequestError:
            self.log.info("Failed to reach node")
            return
            
        for rid in payload.rids:
            if rid == self.identity.rid:
                self.log.info("Skipping myself")
                continue
            if self.cache.exists(rid):
                self.log.info(f"Skipping known RID {rid!r}")
                continue
            
            # marked as external since we are handling RIDs from another node
            # will fetch remotely instead of checking local cache
            self.kobj_queue.push(rid=rid, source=kobj.rid)
        self.log.info("Done")



from dataclasses import dataclass

from rid_lib.types import KoiNetNode

from ..config.base import BaseNodeConfig
from ..protocol.node import NodeProfile


@dataclass
class NodeIdentity:
    """Represents a node's identity (RID, profile)."""
    
    config: BaseNodeConfig
    
    @property
    def rid(self) -> KoiNetNode:
        return self.config.koi_net.node_rid
    
    @property
    def profile(self) -> NodeProfile:
        return self.config.koi_net.node_profile



from pathlib import Path

from .infra import LogSystem
from .base import BaseAssembly
from .config.base import BaseNodeConfig
from .config.full_node import FullNodeConfig
from .config.partial_node import PartialNodeConfig
from .components import (
    Cache,
    Effector,
    Handshaker,
    SyncManager,
    PortManager,
    NodeIdentity,
    KnowledgeProcessingWorker, 
    EventProcessingWorker,
    ErrorHandler,
    EventQueue,
    NetworkGraph,
    RequestHandler,
    NetworkResolver,
    ResponseHandler,
    EventBuffer,
    KnowledgePipeline,
    KobjQueue,
    SecureManager,
    ProfileMonitor,
    NodeServer, 
    NodePoller,
    BasicManifestHandler,
    BasicNetworkOutputFilter,
    BasicRidHandler,
    ForgetNodeHandler,
    NodeContactHandler,
    EdgeNegotiationHandler,
    SecureProfileHandler,
    ConfigProvider
)


class BaseNode(BaseAssembly):
    _log_system: LogSystem = LogSystem
    kobj_queue: KobjQueue = KobjQueue
    event_queue: EventQueue = EventQueue
    poll_event_buf: EventBuffer = EventBuffer
    broadcast_event_buf: EventBuffer = EventBuffer
    config_schema: BaseNodeConfig = BaseNodeConfig
    config: ConfigProvider | BaseNodeConfig = ConfigProvider
    cache: Cache = Cache
    identity: NodeIdentity = NodeIdentity
    graph: NetworkGraph = NetworkGraph
    secure_manager: SecureManager = SecureManager
    handshaker: Handshaker = Handshaker
    error_handler: ErrorHandler = ErrorHandler
    request_handler: RequestHandler = RequestHandler
    sync_manager: SyncManager = SyncManager
    response_handler: ResponseHandler = ResponseHandler
    resolver: NetworkResolver = NetworkResolver
    effector: Effector = Effector
    pipeline: KnowledgePipeline = KnowledgePipeline
    kobj_worker: KnowledgeProcessingWorker = KnowledgeProcessingWorker
    event_worker: EventProcessingWorker = EventProcessingWorker
    profile_monitor: ProfileMonitor = ProfileMonitor
    
    # knowledge handlers
    
    basic_manifest_handler: BasicManifestHandler = BasicManifestHandler
    basic_network_output_filter: BasicNetworkOutputFilter = BasicNetworkOutputFilter
    basic_rid_handler: BasicRidHandler = BasicRidHandler
    forget_node_handler: ForgetNodeHandler = ForgetNodeHandler
    node_contact_handler: NodeContactHandler = NodeContactHandler
    edge_negotiation_handler: EdgeNegotiationHandler = EdgeNegotiationHandler
    secure_profile_handler: SecureProfileHandler = SecureProfileHandler
    
    def __new__(cls, *args, root_dir: Path = Path.cwd(), **kwargs):
        cls._log_system()
        return super().__new__(cls, *args, root_dir=root_dir, **kwargs)

class FullNode(BaseNode):
    config: FullNodeConfig
    server: NodeServer = NodeServer
    port_manager: PortManager = PortManager

class PartialNode(BaseNode):
    config: PartialNodeConfig
    poller: NodePoller = NodePoller



from typing import Any, Self

import structlog

from ..exceptions import BuildError
from .build_artifact import BuildArtifact, CompType
from .container import NodeContainer

log = structlog.stdlib.get_logger()


class Assembler:
    _container: type[NodeContainer] = NodeContainer
    
    # optional order overrides:
    _start_order: list[str]
    _stop_order: list[str]
    
    # annotation hack to show the components and container methods
    def __new__(cls, *args, **kwargs) -> Self | NodeContainer:
        """Returns assembled node container."""
        
        log.debug(f"Assembling '{cls.__name__}'")
        
        artifact = BuildArtifact(cls)
        artifact.build()
        
        components = cls._build_components(artifact)
        
        log.debug("Returning assembled node")
        return cls._container(
            artifact=artifact,
            components=components,
            **kwargs
        )
    
    @staticmethod
    def _build_components(artifact: BuildArtifact):
        """Returns assembled components as a dict."""
        
        log.debug("Building components...")
        components: dict[str, Any] = {}
        for comp_name in artifact.init_order:
        # for comp_name, (comp_type, dep_names) in dep_graph.items():
            comp = artifact.comp_dict[comp_name]
            comp_type = artifact.comp_types[comp_name]
            
            if comp_type == CompType.OBJECT:
                components[comp_name] = comp
            
            elif comp_type == CompType.SINGLETON:
                # builds depedency dict for current component
                dependencies = {}
                for dep in artifact.init_graph[comp_name]:
                    if dep not in components:
                        raise BuildError(f"Couldn't find required component '{dep}'")
                    dependencies[dep] = components[dep]
                components[comp_name] = comp(**dependencies)
        log.debug("Done")
        
        return components



from .env_config import EnvConfig
from .poller_config import PollerConfig
from .server_config import ServerConfig
from .koi_net_config import (
    KoiNetConfig,
    EventWorkerConfig,
    KobjWorkerConfig,
    NodeContact
)
from .full_node import FullNodeConfig, FullNodeProfile
from .partial_node import PartialNodeConfig, PartialNodeProfile
from ..protocol.node import NodeProvides


from pydantic import BaseModel


class ServerConfig(BaseModel):
    """Server config for full nodes.
    
    The parameters in this class represent how a server should be hosted,
    not accessed. For example, a node may host a server at
    `http://127.0.0.1:8000/koi-net`, but serve through nginx at
    `https://example.com/koi-net`.
    """
    
    host: str = "127.0.0.1"
    port: int = 8000
    path: str | None = "/koi-net"
    
    @property
    def url(self) -> str:
        return f"http://{self.host}:{self.port}{self.path or ''}"


from ..protocol.node import NodeProfile, NodeType
from .base import BaseNodeConfig
from .poller_config import PollerConfig


class PartialNodeProfile(NodeProfile):
    """Node profile config class for partial nodes."""
    base_url: str | None = None
    node_type: NodeType = NodeType.PARTIAL

class PartialNodeConfig(BaseNodeConfig):
    """Node config class for partial nodes."""
    poller: PollerConfig = PollerConfig()


import threading
from logging import Logger
from pathlib import Path
from queue import Queue

import structlog
from structlog.contextvars import bound_contextvars

from .infra import Assembler
from .components import LoggingContext


class BaseAssembly(Assembler):
    root_dir: Path
    
    shutdown_signal: threading.Event = threading.Event
    exception_queue: Queue[Exception] = lambda: Queue()

    log: Logger = lambda root_dir: structlog.stdlib.get_logger().bind(log_dir=root_dir)
    logging_context: LoggingContext = LoggingContext
    
    def __new__(cls, *args, root_dir: Path, **kwargs):
        cls.root_dir = root_dir
        with bound_contextvars(log_dir=root_dir):
            return super().__new__(cls, *args, **kwargs)



from dataclasses import dataclass, field
from queue import Queue

from rid_lib.types import KoiNetNode

from koi_net.protocol.event import Event


@dataclass
class QueuedEvent:
    event: Event
    target: KoiNetNode

@dataclass
class EventQueue:
    """Queue for outgoing network events."""
    q: Queue[QueuedEvent] = field(init=False, default_factory=Queue)
    
    def push(self, event: Event, target: KoiNetNode):
        """Pushes event to queue of specified node.
        
        Event will be sent to webhook or poll queue by the event worker
        depending on the node type and edge type of the specified node.
        """
        
        self.q.put(QueuedEvent(target=target, event=event))
    


from dataclasses import dataclass

from koi_net.protocol.knowledge_object import KnowledgeObject
from koi_net.protocol.event import EventType
from ..interfaces import KnowledgeHandler, STOP_CHAIN, HandlerType
from ..identity import NodeIdentity


@dataclass
class BasicRidHandler(KnowledgeHandler):
    identity: NodeIdentity
    
    handler_type = HandlerType.RID
    
    def handle(self, kobj: KnowledgeObject):
        if kobj.rid == self.identity.rid and kobj.source is not None:
            return STOP_CHAIN
        
        if kobj.event_type == EventType.FORGET:
            kobj.normalized_event_type = EventType.FORGET
            return kobj


from dataclasses import dataclass

from rid_lib.types import KoiNetNode

from koi_net.protocol.edge import EdgeProfile
from koi_net.protocol.knowledge_object import KnowledgeObject
from koi_net.protocol.event import EventType
from ..interfaces import KnowledgeHandler, HandlerType
from ..kobj_queue import KobjQueue
from ..graph import NetworkGraph
from ..cache import Cache


@dataclass
class ForgetNodeHandler(KnowledgeHandler):
    cache: Cache
    graph: NetworkGraph
    kobj_queue: KobjQueue
    
    handler_type = HandlerType.Final
    rid_types=(KoiNetNode,)
    
    def handle(self, kobj: KnowledgeObject):
        """Removes edges to forgotten nodes."""
    
        if kobj.normalized_event_type != EventType.FORGET:
            return
        
        for edge_rid in self.graph.get_edges():
            edge_bundle = self.cache.read(edge_rid)
            if not edge_bundle:
                continue
            edge_profile = edge_bundle.validate_contents(EdgeProfile)
            
            if kobj.rid in (edge_profile.source, edge_profile.target):
                self.log.debug("Identified edge with forgotten node")
                self.kobj_queue.push(rid=edge_rid, event_type=EventType.FORGET)


from .cache import Cache
from .effector import Effector
from .event_worker import EventProcessingWorker
from .kobj_worker import KnowledgeProcessingWorker
from .identity import NodeIdentity
from .logging_context import LoggingContext
from .poller import NodePoller
from .server import NodeServer
from .secure_manager import SecureManager
from .kobj_queue import KobjQueue
from .pipeline import KnowledgePipeline
from .error_handler import ErrorHandler
from .event_buffer import EventBuffer
from .event_queue import EventQueue
from .graph import NetworkGraph
from .request_handler import RequestHandler
from .resolver import NetworkResolver
from .response_handler import ResponseHandler
from .handshaker import Handshaker
from .port_manager import PortManager
from .profile_monitor import ProfileMonitor
from .sync_manager import SyncManager
from .config_provider import ConfigProvider

from .knowledge_handlers.basic_rid_handler import BasicRidHandler
from .knowledge_handlers.basic_manifest_handler import BasicManifestHandler
from .knowledge_handlers.edge_negotiation_handler import EdgeNegotiationHandler
from .knowledge_handlers.forget_node_handler import ForgetNodeHandler
from .knowledge_handlers.secure_profile_handler import SecureProfileHandler
from .knowledge_handlers.basic_network_output_filter import BasicNetworkOutputFilter
from .knowledge_handlers.node_contact_handler import NodeContactHandler


from dataclasses import dataclass, field
from logging import Logger
from typing import TYPE_CHECKING

from rid_lib.core import RID, RIDType
from rid_lib.ext import Bundle

if TYPE_CHECKING:
    from ..effector import Effector


@dataclass
class DerefHandler:
    log: Logger
    effector: "Effector"
    
    rid_types: tuple[RIDType] = field(init=False)
    
    def __post_init__(self):
        self.effector.register_handler(self)
    
    def handle(self, rid: RID) -> Bundle | None:
        ...


from pydantic import model_validator

from ..protocol import NodeProfile, NodeType
from .base import BaseNodeConfig
from .server_config import ServerConfig


class FullNodeProfile(NodeProfile):
    """Node profile config class for full nodes."""
    node_type: NodeType = NodeType.FULL

class FullNodeConfig(BaseNodeConfig):
    """Node config class for full nodes."""
    server: ServerConfig = ServerConfig()
    
    @model_validator(mode="after")
    def check_url(self):
        """Generates base URL if missing from node profile."""
        if not self.koi_net.node_profile.base_url:
            self.koi_net.node_profile.base_url = self.server.url
        return self



from .assembler import Assembler
from .build_artifact import BuildArtifact
from .component import provides, depends_on, CompType
from .container import NodeContainer
from .lifecycle import NodeLifecycle, NodeState
from .log_system import LogSystem



import time
from pathlib import Path

from pydantic import BaseModel
from rich.console import Console

from koi_net.config.base import BaseNodeConfig
from koi_net.protocol.node import NodeType
from koi_net.infra import NodeState
from koi_net.components.config_provider import ConfigProvider
from .node import NodeInterface


class KoiNetworkConfig(BaseModel):
    first_contact: str | None = None
    nodes: dict[str, str] = {}

class NetworkConfigProvider(ConfigProvider):
    _file_path: str = "koi-network-config.yaml"

class NetworkInterface:
    config: KoiNetworkConfig | NetworkConfigProvider
    
    def __init__(self):
        self.config_schema = KoiNetworkConfig
        self.config = NetworkConfigProvider(
            config_schema=self.config_schema,
            root_dir=Path.cwd()
        )
        
        self.console = Console()
        
        self.nodes: list[NodeInterface] = []
        self.load_nodes()
    
    def load_nodes(self) -> list[NodeInterface]:
        for name, module in self.config.nodes.items():
            node = NodeInterface.from_ref(name, module)
            self.nodes.append(node)
            
            if node.exists():
                # print(f"Loading node '{node.name}'...")
                node.init()
            
    def resolve_node(self, name: str) -> NodeInterface | None:
        for node in self.nodes:
            if node.name == name:
                return node
            
    def add_node(self, node: NodeInterface):
        self.nodes.append(node)
        self.config.nodes[node.name] = node.module
        self.config.save_to_yaml()
        
        fc_node = self.get_first_contact()
        if fc_node:
            with node.mutate_config() as config:
                self.apply_first_contact(
                    source=fc_node.node.config,
                    target=config
                )
    
    def remove_node(self, node: NodeInterface):
        self.nodes.remove(node)
        
        if node.name in self.config.nodes:
            del self.config.nodes[node.name]
            self.config.save_to_yaml()
            
        if self.config.first_contact == node.name:
            self.unset_first_contact(node)
    
    def get_first_contact(self):
        if not self.config.first_contact:
            return
        
        return self.resolve_node(self.config.first_contact)
    
    def apply_first_contact(self, source: BaseNodeConfig, target: BaseNodeConfig):
        target.koi_net.first_contact.rid = source.koi_net.node_rid
        target.koi_net.first_contact.url = source.koi_net.node_profile.base_url
    
    def set_first_contact(self, fc_node: NodeInterface):
        print("Setting first contact...")
        if fc_node.node.config.koi_net.node_profile.node_type == NodeType.PARTIAL:
            print("Partial nodes cannot be first contacts")
            return
        
        prev_fc_node = self.get_first_contact()
        if prev_fc_node:
            if prev_fc_node == fc_node:
                print(f"First contact is already {fc_node.name}")
                return
            self.unset_first_contact(prev_fc_node)
        
        self.config.first_contact = fc_node.name
        self.config.save_to_yaml()
        
        mutated_nodes: int = 0
        for node in self.nodes:
            if node is fc_node:
                continue
            
            with node.mutate_config() as config:
                self.apply_first_contact(
                    source=fc_node.node.config,
                    target=config
                )
                mutated_nodes += 1
        
        print(f"Updated configuration of {mutated_nodes} node(s)")
    
    def unset_first_contact(self, fc_node: NodeInterface):
        print("Unsetting first contact...")
        if not self.config.first_contact:
            return
        
        self.config.first_contact = None
        self.config.save_to_yaml()
        
        mutated_nodes: int = 0
        for node in self.nodes:
            with node.mutate_config() as config:
                if config.koi_net.first_contact.rid == fc_node.node.config.koi_net.node_rid:
                    config.koi_net.first_contact.rid = None
                    config.koi_net.first_contact.url = None
                    mutated_nodes += 1
        
        print(f"Updated configuration of {mutated_nodes} node(s)")

    def sync(self):
        fc_node = self.get_first_contact()
        if fc_node and not fc_node.exists():
            fc_node.create()
            
        for node in self.nodes:
            if not node.exists():
                node.create()
            
            if fc_node and node is not fc_node:
                with node.mutate_config() as config:
                    self.apply_first_contact(
                        source=fc_node.node.config,
                        target=config
                    )
    
    def wipe_cache(self):
        for node in self.nodes:
            if node.exists():
                node.wipe_cache()
                
    def wipe_logs(self):
        for node in self.nodes:
            if node.exists():
                node.wipe_logs()
    
    def wipe_config(self):
        for node in self.nodes:
            if node.exists():
                node.wipe_config()

    def run(self):
        try:
            self.start()
            
            print(f"Completed startup of {len(self.nodes)} node(s)!")
            print("Press Ctrl + C to quit")
            while any(n.state() is NodeState.RUNNING for n in self.nodes):
                time.sleep(0.5)
                
        except KeyboardInterrupt:
            pass
        
        finally:
            self.stop()
    
    def start(self, block: bool = True):
        for name in self.config.nodes:
            node = self.resolve_node(name)
            if node.state() == NodeState.IDLE:
                node.start(block=block)
            
    def stop(self, block: bool = True):
        for name in reversed(self.config.nodes):
            node = self.resolve_node(name)
            if node.state() == NodeState.RUNNING:
                node.stop(block=block)
        



from .node import (
    NodeType,
    NodeProvides,
    NodeProfile
)
from .edge import (
    EdgeStatus,
    EdgeType,
    EdgeProfile,
    generate_edge_bundle
)
from .envelope import SignedEnvelope, UnsignedEnvelope
from .errors import ErrorType, EXCEPTION_TO_ERROR_TYPE
from .event import Event, EventType
from .knowledge_object import KnowledgeObject
from .secure import PublicKey, PrivateKey



from dataclasses import dataclass

from rid_lib.types import KoiNetEdge, KoiNetNode

from koi_net.protocol.edge import EdgeProfile
from koi_net.protocol.knowledge_object import KnowledgeObject
from ..interfaces import KnowledgeHandler, HandlerType
from ..identity import NodeIdentity
from ..graph import NetworkGraph


@dataclass
class BasicNetworkOutputFilter(KnowledgeHandler):
    identity: NodeIdentity
    graph: NetworkGraph
    
    handler_type = HandlerType.Network
    
    def handle(self, kobj: KnowledgeObject):
        """Sets network targets of outgoing event for knowledge object.
        
        Allows broadcasting of all RID types this node is an event provider 
        for (set in node profile), and other nodes have subscribed to. All 
        nodes will also broadcast events about their own (internally sourced) 
        KOI node, and KOI edges that they are part of, regardless of their 
        node profile configuration. Finally, nodes will also broadcast about 
        edges to the other node involved (regardless of if they are subscribed).
        """
        
        involves_this_node = False
        # internally source knowledge objects
        if kobj.source is None:
            if type(kobj.rid) is KoiNetNode:
                if (kobj.rid == self.identity.rid):
                    involves_this_node = True
            
            elif type(kobj.rid) is KoiNetEdge:
                edge_profile = kobj.bundle.validate_contents(EdgeProfile)
                
                if edge_profile.source == self.identity.rid:
                    self.log.debug(f"Adding edge target '{edge_profile.target!r}' to network targets")
                    kobj.network_targets.add(edge_profile.target)
                    involves_this_node = True
                    
                elif edge_profile.target == self.identity.rid:
                    self.log.debug(f"Adding edge source '{edge_profile.source!r}' to network targets")
                    kobj.network_targets.add(edge_profile.source)
                    involves_this_node = True
        
        if (type(kobj.rid) in self.identity.profile.provides.event or involves_this_node):
            subscribers = self.graph.get_neighbors(
                direction="out",
                allowed_type=type(kobj.rid)
            )
            
            self.log.debug(f"Updating network targets with '{type(kobj.rid)}' subscribers: {subscribers}")
            kobj.network_targets.update(subscribers)
            
        if kobj.source and kobj.source in kobj.network_targets:
            kobj.network_targets.remove(kobj.source)
            self.log.debug(f"Removed event source '{kobj.source}' from network targest")
            
        return kobj


import threading
from dataclasses import dataclass, field
from queue import Empty, Queue
from logging import Logger
from enum import StrEnum
from typing import TYPE_CHECKING, Any

from rich.console import Console
from rich.traceback import Traceback

from .build_artifact import BuildArtifact
from .consts import START_FUNC_NAME, STOP_FUNC_NAME

if TYPE_CHECKING:
    from koi_net.components import LoggingContext


class NodeState(StrEnum):
    IDLE = "IDLE"
    STARTING = "STARTING"
    RUNNING = "RUNNING"
    STOPPING = "STOPPING"

@dataclass
class NodeLifecycle:
    log: Logger
    shutdown_signal: threading.Event
    exception_queue: Queue[Exception]
    logging_context: "LoggingContext"
    artifact: BuildArtifact
    container: Any
    
    state: NodeState = field(init=False, default=NodeState.IDLE)
    thread: threading.Thread | None = field(init=False, default=None)
    startup_signal: threading.Event = field(init=False, default_factory=threading.Event)
    
    def start(self, block: bool = True):
        """Starts the lifecycle manager thread, beginning node startup."""
        
        if self.state != NodeState.IDLE:
            self.log.warning("Node can't be started from non-idle state")
            return
        
        self.startup_signal.clear()
        self.thread = threading.Thread(target=self._run)
        self.thread.start()
        
        if block:
            self.startup_signal.wait()
        
    def stop(self, block: bool = True):
        """Signals to lifecycle manager thread, beginning node shutdown."""
        
        if self.state != NodeState.RUNNING:
            self.log.warning("Node can't be stopped from non-running state")
            return
        
        self.shutdown_signal.set()
        
        if block and self.thread and self.thread.is_alive():
            self.thread.join()

    def _run(self):
        with self.logging_context.bound_vars(thread=self.__class__.__name__):
            try:
                self._startup()
                self.startup_signal.set()
                self.shutdown_signal.wait()
                
            finally:
                self._shutdown()
                
                while True:
                    try:
                        exc = self.exception_queue.get_nowait()
                        traceback = Traceback.from_exception(
                            exc_type=type(exc),
                            exc_value=exc,
                            traceback=exc.__traceback__
                        )
                        print()
                        Console().print(traceback)
                        
                    except Empty:
                        break
        
    def _startup(self):
        self.state = NodeState.STARTING
        self.log.info("Starting node...")
        for comp_name in self.artifact.start_order:
            comp = getattr(self.container, comp_name)
            start_func = getattr(comp, START_FUNC_NAME)
            self.log.info(f"Starting {comp_name}...")
            
            try:
                start_func()
            except Exception as err:
                print()
                self.log.error("Startup error: " + str(err))
                self.exception_queue.put(err)
                self.shutdown_signal.set()
            
            if self.shutdown_signal.is_set():
                self.log.error(f"Startup failed, aborting")
                return
        
        self.state = NodeState.RUNNING
        self.log.info("Startup complete!")
            
    def _shutdown(self):
        self.state = NodeState.STOPPING
        self.log.info("Stopping node...")
        for comp_name in self.artifact.stop_order:
            comp = getattr(self.container, comp_name)
            stop_func = getattr(comp, STOP_FUNC_NAME)
            self.log.info(f"Stopping {comp_name}...")
            
            try:
                stop_func()
            except Exception as err:
                self.log.error("Shutdown error:", str(err))
                self.exception_queue.put(err)
        
        self.shutdown_signal.clear()
        self.state = NodeState.IDLE
        self.log.info("Shutdown complete!")




from dataclasses import dataclass

from rid_lib.types import KoiNetNode
from rid_lib.ext.utils import sha256_hash

from koi_net.protocol.knowledge_object import KnowledgeObject
from koi_net.protocol.event import EventType
from koi_net.protocol.node import NodeProfile
from ..identity import NodeIdentity
from ..interfaces import KnowledgeHandler, STOP_CHAIN, HandlerType


@dataclass
class SecureProfileHandler(KnowledgeHandler):
    identity: NodeIdentity
    
    handler_type = HandlerType.Bundle
    rid_types = (KoiNetNode,)
    event_types = (EventType.NEW, EventType.UPDATE)
    
    def handle(self, kobj: KnowledgeObject):
        node_profile = kobj.bundle.validate_contents(NodeProfile)
        node_rid: KoiNetNode = kobj.rid
        
        if sha256_hash(node_profile.public_key) != node_rid.hash:
            self.log.warning(f"Public key hash mismatch for {node_rid!r}!")
            return STOP_CHAIN
        
        if (node_rid != self.identity.rid) and (node_profile.base_url == self.identity.profile.base_url):
            self.log.warning(f"Ignoring node claiming same base URL: {node_profile.base_url}")
            return STOP_CHAIN



from dataclasses import dataclass, field
from enum import StrEnum
from logging import Logger
from typing import TYPE_CHECKING
from rid_lib.core import RIDType

from koi_net.protocol.event import EventType
from koi_net.protocol.knowledge_object import KnowledgeObject

if TYPE_CHECKING:
    from ..pipeline import KnowledgePipeline


class StopChain:
    """Class for STOP_CHAIN sentinel returned by knowledge handlers."""
    pass

STOP_CHAIN = StopChain()

class HandlerType(StrEnum):
    """Types of handlers used in knowledge processing pipeline.
    
    - RID - provided RID; if event type is `FORGET`, this handler decides whether to delete the knowledge from the cache by setting the normalized event type to `FORGET`, otherwise this handler decides whether to validate the manifest (and fetch it if not provided).
    - Manifest - provided RID, manifest; decides whether to validate the bundle (and fetch it if not provided).
    - Bundle - provided RID, manifest, contents (bundle); decides whether to write knowledge to the cache by setting the normalized event type to `NEW` or `UPDATE`.
    - Network - provided RID, manifest, contents (bundle); decides which nodes (if any) to broadcast an event about this knowledge to. (Note, if event type is `FORGET`, the manifest and contents will be retrieved from the local cache, and indicate the last state of the knowledge before it was deleted.)
    - Final - provided RID, manifests, contents (bundle); final action taken after network broadcast.
    """
    
    RID = "rid", 
    Manifest = "manifest", 
    Bundle = "bundle", 
    Network = "network", 
    Final = "final"

@dataclass
class KnowledgeHandler:
    """Handles knowledge processing events of the provided types."""
    
    log: Logger
    pipeline: "KnowledgePipeline"
    
    handler_type: HandlerType = field(init=False)
    rid_types: tuple[RIDType] = field(init=False, default=())
    event_types: tuple[EventType | None] = field(init=False, default=())
    
    def __post_init__(self):
        self.pipeline.register_handler(self)
    
    def handle(self, kobj: KnowledgeObject) -> KnowledgeObject | None | StopChain:
        ...



from dataclasses import dataclass

from rid_lib.ext import Bundle
from rid_lib.types import KoiNetEdge, KoiNetNode

from koi_net.protocol.node import NodeProfile, NodeType
from koi_net.protocol.edge import EdgeProfile, EdgeStatus, EdgeType
from koi_net.protocol.knowledge_object import KnowledgeObject
from koi_net.protocol.event import Event, EventType
from ..interfaces import KnowledgeHandler, STOP_CHAIN, HandlerType
from ..identity import NodeIdentity
from ..kobj_queue import KobjQueue
from ..event_queue import EventQueue
from ..cache import Cache


@dataclass
class EdgeNegotiationHandler(KnowledgeHandler):
    identity: NodeIdentity
    cache: Cache
    event_queue: EventQueue
    kobj_queue: KobjQueue
    
    handler_type = HandlerType.Bundle
    rid_types = (KoiNetEdge,)
    event_types = (EventType.NEW, EventType.UPDATE)
    
    def handle(self, kobj: KnowledgeObject):
        """Handles edge negotiation process.
        
        Automatically approves proposed edges if they request RID types this 
        node can provide (or KOI node, edge RIDs). Validates the edge type 
        is allowed for the node type (partial nodes cannot use webhooks). If 
        edge is invalid, a `FORGET` event is sent to the other node.
        """

        # only handle incoming events (ignore internal edge knowledge objects)
        if kobj.source is None: 
            return
        
        edge_profile = kobj.bundle.validate_contents(EdgeProfile)

        # indicates peer subscribing to this node
        if edge_profile.source == self.identity.rid:
            if edge_profile.status != EdgeStatus.PROPOSED:
                return
            
            self.log.debug("Handling edge negotiation")
            
            peer_rid = edge_profile.target
            peer_bundle = self.cache.read(peer_rid)
            
            if not peer_bundle:
                self.log.warning(f"Peer {peer_rid!r} unknown to me")
                return STOP_CHAIN
            
            peer_profile = peer_bundle.validate_contents(NodeProfile)
            
            # explicitly provided event RID types and (self) node + edge objects
            provided_events = (
                *self.identity.profile.provides.event,
                KoiNetNode, KoiNetEdge
            )
            
            abort = False
            if (edge_profile.edge_type == EdgeType.WEBHOOK and 
                peer_profile.node_type == NodeType.PARTIAL):
                self.log.debug("Partial nodes cannot use webhooks")
                abort = True
            
            if not set(edge_profile.rid_types).issubset(provided_events):
                not_provided = set(edge_profile.rid_types) - set(provided_events)
                self.log.debug(f"Requested RID types {not_provided} not provided by this node")
                abort = True
            
            if abort:
                event = Event.from_rid(EventType.FORGET, kobj.rid)
                self.event_queue.push(event, peer_rid)
                return STOP_CHAIN
            
            else:
                self.log.debug("Approving proposed edge")
                edge_profile.status = EdgeStatus.APPROVED
                updated_bundle = Bundle.generate(kobj.rid, edge_profile.model_dump())

                self.kobj_queue.push(bundle=updated_bundle, event_type=EventType.UPDATE)
                return
                
        elif edge_profile.target == self.identity.rid:
            if edge_profile.status == EdgeStatus.APPROVED:
                self.log.debug("Edge approved by other node!")

